/******************************************************************************

JMixNet - A Java MixNet Implementation

Fabiano Castro Pereira
LabSEC - Computer Security Laboratory (www.labsec.ufsc.br)
Informatics and Statistics Department (www.inf.ufsc.br)
Federal University of Santa Catarina  (www.ufsc.br)  
Florianopolis, Santa Catarina, Brazil

JMixNet is a registered trademark of the Federal University of Santa Catarina.
No commercial use of this trademark may be made without prior written
permission of the Federal University of Santa Catarina.\

******************************************************************************/
package br.edu.ufsc.labsec.crypto;

import java.security.*;
import java.security.cert.*;
import java.security.cert.Certificate;
import java.io.*;
import javax.crypto.*;

/**
 * @author castro
 *
 * Class to deal with asymmetric cryptographic operations.
 */
public class AsymCryptographer {
    
    protected KeyStore keyStore;
    protected char[] keyStorePassword;
    protected Cipher cipher;
    protected Key privateKey = null;
    protected Certificate certificate = null;
    
    protected static final int OP_NONE    = 0;
    protected static final int OP_ENCRYPT = 1;
    protected static final int OP_DECRYPT = 2;
    protected static final int OP_SIGN    = 3;
    protected static final int OP_VERIFY  = 4;
    protected int currentOperation = OP_NONE;
    
    protected String transformation = "RSA";
    
    /**
     * Constructs a new asymmetric cryptographer using the keystore information provided.  
     * 
     * @param keyStoreFilename Name of the file containing the keystore. 
     * @param keyStorePassword Password to open the keystore.
     * 
     * @throws CryptoException if some unexpected error occurred.
     */
    public AsymCryptographer(String keyStoreFilename, String keyStorePass) throws CryptoException {
        try {
            keyStore = KeyStore.getInstance("JKS");
            keyStorePassword = keyStorePass.toCharArray();
            keyStore.load(new FileInputStream(keyStoreFilename), keyStorePassword);
            
            cipher = Cipher.getInstance(transformation);
        }
        catch (KeyStoreException ex) {
            throw new CryptoException("Error opening keystore file.", ex);
        }
        catch (IOException ex) {
            throw new CryptoException("Error reading keystore file.", ex);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new CryptoException("Error with algorithm definitions.", ex);
        }
        catch (CertificateException ex) {
            throw new CryptoException("Error loading certificate from keystore.", ex);
        }
        catch (NoSuchPaddingException ex) {
            throw new CryptoException("Error with padding definitions.", ex);
        }
    }
    
    /**
     * Constructs a new asymmetric cryptographer using the keystore information provided.  
     * 
     * @throws CryptoException if some unexpected error occurred.
     */
    public AsymCryptographer() throws CryptoException {
        try {
            cipher = Cipher.getInstance(transformation);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new CryptoException("Error with algorithm definitions.", ex);
        }
        catch (NoSuchPaddingException ex) {
            throw new CryptoException("Error with padding definitions.", ex);
        }
    }
    
    /**
     * Tells the criptographer to use the desired alias content.
     * 
     * @param newAlias The alias to be used hereafter.
     * @param password The alias password. Null to use the keystore password.
     * @throws CryptoException If some keystore error occurred, or a bad password was informed.
     */
    public void useAlias(String newAlias, String password) throws CryptoException {
        try {
            //verify and use the private key contained in alias             
            if (keyStore.isKeyEntry(newAlias)) {
                if (password == null) {
                    privateKey = keyStore.getKey(newAlias, keyStorePassword);
                }
                else {
                    privateKey = keyStore.getKey(newAlias, password.toCharArray());
                }
            }
            //use the certificate contained in alias             
            certificate = keyStore.getCertificate(newAlias);
            currentOperation = OP_NONE;
        }
        catch (KeyStoreException ex) {
            throw new CryptoException("Error using alias.", ex);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new CryptoException("Error using alias.", ex);
        }
        catch (UnrecoverableKeyException ex) {
            throw new CryptoException("Error using alias, maybe password is wrong.", ex);
        }
    }
    
    /**
     * Sets the certificate to be used.
     * 
     * @param cert The certificate to be used.
     */
    public void setCertificate(Certificate cert) {
        certificate = cert;
        currentOperation = OP_NONE;
    }
    
    /**
     * Gets the certificate in use.
     * 
     * @return The certificate in use.
     */
    public Certificate getCertificate() {
        return certificate;
    }
    
    /**
     * Verify if the cipher is initialized with the desired operation.
     * 
     * @param operation Desired operation
     * @throws CryptoException If an invalid key or algorithm parameter is found.
     */
    protected void verifyOperation(int operation) throws CryptoException {
        try {
            switch (operation) {
                case OP_ENCRYPT :
                    if (currentOperation != OP_ENCRYPT) {
                        currentOperation = OP_ENCRYPT;
                        cipher.init(Cipher.ENCRYPT_MODE, certificate);
                    }
                    break;
                case OP_DECRYPT :
                    if (currentOperation != OP_DECRYPT) {
                        currentOperation = OP_DECRYPT;
                        cipher.init(Cipher.DECRYPT_MODE, privateKey);
                    }
                    break;
            }
        }
        catch (InvalidKeyException ex) {
            throw new CryptoException("Invalid symmetric key.", ex);
        } 
    }
    
    /**
     * Returns the desired plain data encrypted using the certificate defined previously.
     * 
     * @param plainData the data to be encrypted.
     * @return the encrypted data.
     * @throws CryptoException if a padding or block size error occurred.
     */
    public byte[] encrypt(byte[] plainData) throws CryptoException {
        try {
            verifyOperation(OP_ENCRYPT);
            return cipher.doFinal(plainData);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Error with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Error with data block size.", ex);
        }
    }
    
    /**
     * Returns the desired plain data encrypted using the certificate defined previously.
     * 
     * @param plainData the data to be encrypted.
     * @param offset the offset in <tt>plainData</tt> where the plain data starts.
     * @param length the plain data length.
     * @param output the byte buffer to write the encrypted data.
     * @return the amount of encrypted data.
     * @throws CryptoException if some error occurred.
     */
    public int encrypt(byte[] plainData, int offset, int length, byte[] output) throws CryptoException {
        try {
            verifyOperation(OP_ENCRYPT);
            return cipher.doFinal(plainData, offset, length, output);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Error with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Error with data block size.", ex);
        }
        catch (ShortBufferException ex) {
            throw new CryptoException("Buffer too small.", ex);
        }
    }
    
    /**
     * Returns the desired plain data encrypted using the certificate defined previously.
     * 
     * @param plainData the data to be encrypted.
     * @param offset the offset in <tt>plainData</tt> where the plain data starts.
     * @param length the plain data length.
     * @param output the byte buffer to write the encrypted data.
     * @param outputOffset the offset in <tt>output</tt> where the encrypted data starts.
     * @return the amount of encrypted data.
     * @throws CryptoException if some error occurred.
     */
    public int encrypt(byte[] plainData, int offset, int length, byte[] output, int outputOffset) throws CryptoException {
        try {
            verifyOperation(OP_ENCRYPT);
            return cipher.doFinal(plainData, offset, length, output, outputOffset);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Error with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Error with data block size.", ex);
        }
        catch (ShortBufferException ex) {
            throw new CryptoException("Buffer too small.", ex);
        }
    }
    
    /**
     * Returns the desired encrypted data decrypted using the private key defined previously.
     * 
     * @param encryptedData the data to be decrypted.
     * @return the decrypted data.
     * @throws CryptoException if a padding or block size error occurred.
     */
    public byte[] decrypt(byte[] encryptedData) throws CryptoException {
        try {
            verifyOperation(OP_DECRYPT);
            return cipher.doFinal(encryptedData);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Error with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Error with data block size.", ex);
        }
    }
    
    /**
     * Returns the desired encrypted data decrypted using the private key defined previously.
     * 
     * @param encryptedData the data to be decrypted.
     * @param offset the offset in <tt>encryptedData</tt> where the encrypted data starts.
     * @param length the encrypted data length.
     * @return the decrypted data.
     * @throws CryptoException if a padding or block size error occurred.
     */
    public byte[] decrypt(byte[] encryptedData, int offset, int length) throws CryptoException {
        try {
            verifyOperation(OP_DECRYPT);
            return cipher.doFinal(encryptedData, offset, length);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Error with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Error with data block size.", ex);
        }
    }
    
    /**
     * Returns the desired encrypted data decrypted using the private key defined previously.
     * 
     * @param encryptedData the data to be decrypted.
     * @param offset the offset in <tt>encryptedData</tt> where the encrypted data starts.
     * @param length the encrypted data length.
     * @param output the byte buffer to write the decrypted data.
     * @return the amount of decrypted data.
     * @throws CryptoException if a padding or block size error occurred.
     */
    public int decrypt(byte[] encryptedData, int offset, int length, byte[] output) throws CryptoException {
        try {
            verifyOperation(OP_DECRYPT);
            return cipher.doFinal(encryptedData, offset, length, output);
        }
        catch (IllegalArgumentException ex) {
            throw new CryptoException("Error with encrypted data.", ex);
        }        
        catch (BadPaddingException ex) {
            throw new CryptoException("Error with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Error with data block size.", ex);
        }
        catch (IllegalStateException ex) {
            throw new CryptoException("Wrong state found.", ex);
        } 
        catch (ShortBufferException ex) {
            throw new CryptoException("Buffer too small.", ex);
        }
    }
    
    /**
     * Continues a multiple-part encryption or decryption operation (depending on how this cipher was initialized), 
     * processing another data part.
     * 
     * @param input the input buffer
     * @param inputOffset the offset in <tt>input</tt> where the input starts
     * @param inputLen the input length
     * @param output the buffer for the result
     * @param outputOffset the <tt>offset</tt> in output where the result is stored
     * @return the number of bytes stored in <tt>output</tt>
     * @throws CryptoException if this cipher is in a wrong state or the given output buffer is too small to hold the result.
     */
    public int addDataToEncrypt(byte[] input, int inputOffset, int inputLen, byte[] output, int outputOffset) throws CryptoException {
        try {
            verifyOperation(OP_ENCRYPT);
            return cipher.update(input, inputOffset, inputLen, output, outputOffset);
        }
        catch (IllegalStateException ex) {
            throw new CryptoException("Wrong state found.", ex);
        }
        catch (ShortBufferException ex) {
            throw new CryptoException("Buffer too small.", ex);
        }
    }
    
    /**
     * @return The asymmetric transformation in use.
     */
    public String getTransformation() {
        return transformation;
    }

    /**
     * @param string The asymmetric transformation to be used.
     */
    public void setTransformation(String string) throws NoSuchPaddingException, NoSuchAlgorithmException {
        transformation = string;
        cipher = Cipher.getInstance(transformation);
        currentOperation = OP_NONE;
    }
}