/******************************************************************************

JMixNet - A Java MixNet Implementation

Fabiano Castro Pereira
LabSEC - Computer Security Laboratory (www.labsec.ufsc.br)
Informatics and Statistics Department (www.inf.ufsc.br)
Federal University of Santa Catarina  (www.ufsc.br)  
Florianopolis, Santa Catarina, Brazil

JMixNet is a registered trademark of the Federal University of Santa Catarina.
No commercial use of this trademark may be made without prior written
permission of the Federal University of Santa Catarina.\

******************************************************************************/
package br.edu.ufsc.labsec.crypto;

import java.io.IOException;
import java.security.*;

import javax.crypto.*;
import javax.crypto.spec.*;

/**
 * @author castro
 *
 * Class to deal with symmetric cryptographic operations.
 */
public class SymCryptographer {
    
    protected Cipher cipher;
    protected String algorithm = "AES";
    protected String transformation = "AES/CBC/NoPadding";

    protected KeyGenerator keyGenerator;
    protected SecretKey symKey;
    protected AlgorithmParameters parameters;
    protected boolean hasParameters = false;
    
    protected static final int OP_NONE    = 0;
    protected static final int OP_ENCRYPT = 1;
    protected static final int OP_DECRYPT = 2;
    protected static final int OP_SIGN    = 3;
    protected static final int OP_VERIFY  = 4;
    protected int currentOperation = OP_NONE;
    
    /**
     * Constructs a new symmetric cryptographer.  
     */
    public SymCryptographer() throws CryptoException {
        try {
            cipher = Cipher.getInstance(transformation);
            keyGenerator = KeyGenerator.getInstance(algorithm);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new CryptoException("Error with algorithm definitions.", ex);
        }
        catch (NoSuchPaddingException ex) {
            throw new CryptoException("Error with padding definitions.", ex);
        }
    }
    
    /**
     * Generates a new ramdom key.
     */
    public void generateKey() {
        symKey = keyGenerator.generateKey();
        currentOperation = OP_NONE;
    }
    
    /**
     * Gets the key material of the symmetric key.
     * 
     * @return The key material of the symmetric key.
     */
    public byte[] getEncodedKey() {
        return symKey.getEncoded();
    }
    
    /**
     * Sets the key material of the symmetric key.
     * 
     * @param key key material.
     * @param offset offset of <tt>key</tt> where key material starts.
     * @param len key material length.
     */
    public void setEncodedKey(byte[] key, int offset, int len) {
        symKey = new SecretKeySpec(key, offset, len, algorithm);
        currentOperation = OP_NONE;
    }
    
    /**
     * Set the cipher parameters to be used.
     * 
     * @param params The encoded parameters.
     * @param offset The offset in <tt>params</tt> where they start. 
     * @param len The parameters length.
     * @throws CryptoException If a parameter initialization error occurred.
     */
    public void setEncodedParameters(byte[] params, int offset, int len) throws CryptoException {
        try {
            byte[] newParams = new byte[len];
            System.arraycopy(params, offset, newParams, 0, len);
            parameters = AlgorithmParameters.getInstance(algorithm);
            parameters.init(newParams);
            hasParameters = true;
            currentOperation = OP_NONE;
        }
        catch (IOException ex) {
            throw new CryptoException("Erro with parameter initialization.", ex);
        } 
        catch (NoSuchAlgorithmException ex) {
            throw new CryptoException("Error with algorithm definitions.", ex);
        }
    }
    
    /**
     * Set the cipher parameters to be used.
     * 
     * @param params The encoded parameters.
     * @throws CryptoException If a parameter initialization error occurred.
     */
    public void setEncodedParameters(byte[] params) throws CryptoException {
        try {
            parameters = AlgorithmParameters.getInstance(algorithm);
            parameters.init(params);
            hasParameters = true;
            currentOperation = OP_NONE;
        }
        catch (IOException ex) {
            throw new CryptoException("Erro with parameter initialization.", ex);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new CryptoException("Error with algorithm definitions.", ex);
        }
    }
    
    /**
     * Discard the cipher parameters in use.
     */
    public void resetParameters() {
        hasParameters = false;
    }
    
    /**
     * Returns the parameters used with this cipher.
     * 
     * @return the parameters used with this cipher, or null if this cipher does not use any parameters.
     */
    public AlgorithmParameters getParameters() {
        return cipher.getParameters();
    }
    
    /**
     * Verify if the cipher is initialized with the desired operation.
     * 
     * @param operation Desired operation
     * @throws CryptoException If an invalid key or algorithm parameter is found.
     */
    protected void verifyOperation(int operation) throws CryptoException {
        try {
            switch (operation) {
                case OP_ENCRYPT :
                    if (currentOperation != OP_ENCRYPT) {
                        currentOperation = OP_ENCRYPT;
                        if (hasParameters) 
                            cipher.init(Cipher.ENCRYPT_MODE, symKey, parameters);
                        else
                            cipher.init(Cipher.ENCRYPT_MODE, symKey);
                    }
                    break;
                case OP_DECRYPT :
                    if (currentOperation != OP_DECRYPT) {
                        currentOperation = OP_DECRYPT;
                        if (hasParameters) 
                            cipher.init(Cipher.DECRYPT_MODE, symKey, parameters);
                        else
                            cipher.init(Cipher.DECRYPT_MODE, symKey);
                    }
                    break;
            }
        }
        catch (InvalidKeyException ex) {
            throw new CryptoException("Invalid symmetric key.", ex);
        } 
        catch (InvalidAlgorithmParameterException ex) {
            throw new CryptoException("Invalid algorithm parameter.", ex);
        } 
    }
    
    /**
     * Returns the desired plain data encrypted.
     * 
     * @param plainData the data to be encrypted.
     * @param inputOffset
     * @param inputLen
     * @param output
     * @param outputOffset
     * @return the encrypted data.
     * @throws CryptoException if some error occurred.
     */
    public int encrypt(byte[] plainData, int inputOffset, int inputLen, byte[] output, int outputOffset) throws CryptoException {
        try {
            verifyOperation(OP_ENCRYPT);
            return cipher.doFinal(plainData, inputOffset, inputLen, output, outputOffset);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Erro with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Erro with data block size.", ex);
        }
        catch (IllegalStateException ex) {
            throw new CryptoException("Wrong state found.", ex);
        } 
        catch (ShortBufferException ex) {
            throw new CryptoException("Buffer too small.", ex);
        }
    }
    
    /**
     * Returns the desired plain data encrypted.
     * 
     * @param plainData the data to be encrypted.
     * @return the encrypted data.
     * @throws CryptoException if a padding or block size error occurred.
     */
    public byte[] encrypt(byte[] plainData) throws CryptoException {
        try {
            verifyOperation(OP_ENCRYPT);
            return cipher.doFinal(plainData);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Erro with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Erro with data block size.", ex);
        }
    }
    
    /**
     * Returns the desired encrypted data decrypted.
     * 
     * @param encryptedData the data to be decrypted.
     * @return the decrypted data.
     * @throws CryptoException if a padding or block size error occurred.
     */
    public byte[] decrypt(byte[] encryptedData) throws CryptoException {
        try {
            verifyOperation(OP_DECRYPT);
            return cipher.doFinal(encryptedData);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Erro with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Erro with data block size.", ex);
        }
    }

    /**
     * Returns the desired encrypted data decrypted.
     * 
     * @param encryptedData the data to be decrypted.
     * @param offset the offset in <tt>encryptedData</tt> where the encrypted data starts.
     * @param length the encrypted data length.
     * @return the decrypted data.
     * @throws CryptoException if a padding or block size error occurred.
     */
    public byte[] decrypt(byte[] encryptedData, int offset, int length) throws CryptoException {
        try {
            verifyOperation(OP_DECRYPT);
            return cipher.doFinal(encryptedData, offset, length);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Erro with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Erro with data block size.", ex);
        }
    }

    /**
     * Returns the desired encrypted data decrypted.
     * 
     * @param encryptedData the data to be decrypted.
     * @param offset the offset in <tt>encryptedData</tt> where the encrypted data starts.
     * @param length the encrypted data length.
     * @param output the byte buffer to write the decrypted data.
     * @return the amount of decrypted data.
     * @throws CryptoException if a padding or block size error occurred.
     */
    public int decrypt(byte[] encryptedData, int offset, int length, byte[] output) throws CryptoException {
        try {
            verifyOperation(OP_DECRYPT);
            return cipher.doFinal(encryptedData, offset, length, output);
        }
        catch (BadPaddingException ex) {
            throw new CryptoException("Erro with data padding.", ex);
        }
        catch (IllegalBlockSizeException ex) {
            throw new CryptoException("Erro with data block size.", ex);
        }
        catch (IllegalStateException ex) {
            throw new CryptoException("Wrong state found.", ex);
        } 
        catch (ShortBufferException ex) {
            throw new CryptoException("Buffer too small.", ex);
        }
    }

    /**
     * @return The transformation in use.
     */
    public String getTransformation() {
        return transformation;
    }

    /**
     * @param string The transformation to be used.
     */
    public void setTransformation(String string) throws NoSuchPaddingException, NoSuchAlgorithmException {
        transformation = string;
        
        //initialize cipher and key generator
        cipher = Cipher.getInstance(transformation);
        currentOperation = OP_NONE;
        algorithm = transformation.split("/")[0];
        keyGenerator = KeyGenerator.getInstance(algorithm);
    }
}