/******************************************************************************

JMixNet - A Java MixNet Implementation

Fabiano Castro Pereira
LabSEC - Computer Security Laboratory (www.labsec.ufsc.br)
Informatics and Statistics Department (www.inf.ufsc.br)
Federal University of Santa Catarina  (www.ufsc.br)  
Florianopolis, Santa Catarina, Brazil

JMixNet is a registered trademark of the Federal University of Santa Catarina.
No commercial use of this trademark may be made without prior written
permission of the Federal University of Santa Catarina.\

******************************************************************************/
package br.edu.ufsc.labsec.jmixnet;

import java.util.*;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.*;

/**
 * @author castro
 *
 * Thread to deliver messages along the mixnet.
 */
public class DelivererThread extends Thread {
    
    //message delivery
    protected LinkedList newMessages;
    protected LinkedList workingMessages;
    protected LinkedList handledMessages;
    protected MessageDeliverer deliverer = null;
    protected OutputStream serverOutput;
    protected boolean deliveryActive = true;
    protected Iterator it;
    protected ByteBuffer currentBuffer;
    
    /**
     * Creates a new deliverer thread to be used by the last server.
     * 
     * @param msgDelivererClass The Class to be used to get a deliverer object.
     * @throws JMixNetException If the deliverer cannot be created. 
     */
    public DelivererThread(String msgDelivererClass) throws JMixNetException {
        try {
            //creates the message deliverer
            deliverer = (MessageDeliverer)(Class.forName(msgDelivererClass)).newInstance();
        } 
        catch (InstantiationException ex) {
            throw new JMixNetException("Unable to instantiate message deliverer.\nDetails: " + ex.getMessage());
        } 
        catch (IllegalAccessException ex) {
            throw new JMixNetException("Unable to instantiate message deliverer.\nDetails: " + ex.getMessage());
        } 
        catch (ClassNotFoundException ex) {
            throw new JMixNetException("Unable to instantiate message deliverer.\nDetails: " + ex.getMessage());
        }
        setupBuffers();
    }
    
    /**
     * Creates a new deliverer thread to be used by a middle server.
     */
    public DelivererThread() {
        setupBuffers();
    }
    
    /**
     * Creates needed buffers.
     */
    protected void setupBuffers() {
        newMessages = new LinkedList();
        handledMessages = new LinkedList();
        workingMessages = new LinkedList();
    }
    
    /**
     * Add new messages to be handled, and return the buffers already handled. 
     * 
     * @param newMsgs New messages to be delivered.
     * @return The buffers already handled.
     */
    public LinkedList exchangeBuffers(LinkedList newMsgs) {
        
        //add received new messages
        synchronized(newMessages) {
            newMessages.addAll(newMsgs);
        }
        newMsgs.clear();
        
        //return handled messages
        synchronized(handledMessages) {
            newMsgs.addAll(handledMessages);
            handledMessages.clear();
        }
        //tell the thread that there are new messages
        interrupt();
        
        return newMsgs;
    }
    
    /**
     * Tells the thread to stop working.
     */
    public void finishDelivery() {
        deliveryActive = false;
        interrupt();        
    }
    
    /**
     * Sets the server output to be used to deliver messages.
     * 
     * @param serverOutput The serverOutput to set.
     */
    public void setServerOutput(OutputStream serverOutput) {
        this.serverOutput = serverOutput;
    }
    
    public void run() {
        
        while (deliveryActive) {
            if (newMessages.size() > 0) {
                //get the new messages
                synchronized(newMessages) {
                    workingMessages.addAll(newMessages);
                    newMessages.clear();
                }
                it = workingMessages.iterator();
                while (it.hasNext()) {
                    currentBuffer = (ByteBuffer)it.next();
                    if (deliverer != null)
                        deliverer.deliverMsg(currentBuffer);
                    else
                        try {
                            serverOutput.write(currentBuffer.array());
                        }
                        catch (IOException ex) {
                            System.err.println("Unable to deliver message to the next server.");
                        }
                    currentBuffer.clear();
                    handledMessages.add(currentBuffer);
                }
                workingMessages.clear();
            } 
            else {
                try {
                    //wait a second to verify new messages 
                    sleep(1000);
                }
                catch (InterruptedException e) {
                    //do nothing
                }
            }        
        }
    }
}