/******************************************************************************

JMixNet - A Java MixNet Implementation

Fabiano Castro Pereira
LabSEC - Computer Security Laboratory (www.labsec.ufsc.br)
Informatics and Statistics Department (www.inf.ufsc.br)
Federal University of Santa Catarina  (www.ufsc.br)  
Florianopolis, Santa Catarina, Brazil

JMixNet is a registered trademark of the Federal University of Santa Catarina.
No commercial use of this trademark may be made without prior written
permission of the Federal University of Santa Catarina.\

******************************************************************************/
package br.edu.ufsc.labsec.jmixnet;

import br.edu.ufsc.labsec.crypto.*;
import br.edu.ufsc.labsec.util.ConfigException;

import java.net.*;
import java.io.*;
import java.nio.ByteBuffer;
import java.util.*;

import java.security.*;
import java.security.cert.*;
import java.security.cert.Certificate;

import org.bouncycastle.jce.provider.BouncyCastleProvider;

/**
 * @author castro
 *
 * Client to send messages through the JMixNet.
 */
public class JMixNetClient {

    //configuration-related
    protected JMixNetConfigManager config;
    protected Iterator iterator;

    //message info
    protected int messageSize;
    protected int sessionKeyFieldSize;
    protected int dataDescSize = 2; //a short (2 bytes) describes the size of meaningful data 
    protected final int sessionKeySize = 16;
    protected final int hashFieldSize = 16; //MD5
    protected byte[] encodedParameters;
    
    //message building
    public int maxDataSize;
    protected ByteBuffer dataBuffer;
    protected ByteBuffer tempBuffer;
    protected byte[] srcBuf;
    protected byte[] dstBuf;
    protected int srcBufContentSize;
    protected int dstBufContentSize;
    
    //reports
    protected PrintStream errorReport = System.err;
    protected PrintStream clientReport = System.out;
    
    //cryptographers and hash
    protected SymCryptographer symCrypto;
    protected AsymCryptographer asymCrypto;
    protected MessageDigest digester;
    
    //to receive and send messages
    protected ArrayList serverCerts;
    protected Socket connection;
    protected InputStream serverInput = null;
    protected OutputStream serverOutput = null;

    /**
     * Construct the client with initial setup.
     */
    public JMixNetClient() throws JMixNetException {
        try {
            config = new JMixNetConfigManager();
        }
        catch (ConfigException ex) {
            showError("Config error occurred:\n" + ex.getMessage());
            throw new JMixNetException();
        }
        configureClient();
    }
        
    /**
     * Construct the client with initial setup, based on specified config file.
     * 
     * @param configFile Configuration file other than "jmixnet.cfg"
     */
    public JMixNetClient(InputStream configStream) throws JMixNetException {
        try {
            config = new JMixNetConfigManager(configStream);
        }
        catch (ConfigException ex) {
            showError("Config error occurred:\n" + ex.getMessage());
            throw new JMixNetException();
        }
        configureClient();
    }
        
    /**
     * Configurations based on config file.
     */
    protected void configureClient() throws JMixNetException {
        
        //ensure Bouncy Castle Provider is available
        if (Security.getProvider("BC") == null) {
            Security.addProvider(new BouncyCastleProvider());
        }
        
        //message config
        messageSize = config.getMessageSize();
        dataBuffer = ByteBuffer.allocate(messageSize);
        tempBuffer = ByteBuffer.allocate(messageSize);
        
        try {
            //server certificates
            serverCerts = (ArrayList)config.getServerCerts();
            Collections.reverse(serverCerts);
        }
        catch (ConfigException ex) {
            showError("Unable to get server certificates:\n" + ex.getMessage());
            throw new JMixNetException(ex);
        }
        //calculate max data size the client can send
        iterator = serverCerts.iterator();
        int totalKeyLength = 0;
        while (iterator.hasNext()) {
            totalKeyLength += X509CertificateInfo.getKeyBytesSize((X509Certificate)iterator.next());
        }
        maxDataSize = messageSize - hashFieldSize - totalKeyLength - dataDescSize;
        
        try {
            //cryptography services and hash
            symCrypto = new SymCryptographer();
            asymCrypto = new AsymCryptographer();
            digester = MessageDigest.getInstance("MD5");
        }
        catch (CryptoException ex) {
            showError("Unable to use cryptography services:\n" + ex.getMessage());
            throw new JMixNetException();
        }
        catch (NoSuchAlgorithmException ex) {
            showError("Unable to get a message digest instance:\n" + ex.getMessage());
            throw new JMixNetException();
        }
    }
    
    /**
     * Uses error report to display desired error message.
     * 
     * @param message Error message to be shown.
     */
    protected void showError(String message) {
        errorReport.println("ERROR: " + message);
    }
    
    /**
     * Uses client report to display desired log message.
     * 
     * @param message Log message to be shown.
     */
    protected void showLog(String message) {
        clientReport.println("Log: " + message);
    }
    
    /**
     * Set the PrintStream object to be used as the error report.
     * 
     * @param report PrintStream object to be used as the error report.
     */
    public void setErrorReport(PrintStream report) {
        errorReport = report;
    }

    /**
     * Set the PrintStream object to be used as the client report.
     * 
     * @param report PrintStream object to be used as the server report.
     */
    public void setClientReport(PrintStream report) {
        clientReport = report;
    }
    
    /**
     * Connect to JMixNet server using configuration data. 
     */
    public void connect() {
        try {
            connection = new Socket(config.getAddress(), config.getPort());
            serverInput = connection.getInputStream(); 
            serverOutput = connection.getOutputStream(); 
        }
        catch (UnknownHostException ex) {
            showError("Unknown host: " + connection.toString() + ".\nDetails: " + ex.getMessage());
        }
        catch (IOException ex) {
            showError("Unable to communicate with JMixNet server.\nDetails: " + ex.getMessage());
        }
    }
    
    /**
     * Disconnect from JMixNet server. 
     */
    public void disconnect() {
        try {
            connection.close();
        }
        catch (IOException ex) {
            showError("Unable to disconnect from JMixNet server.\nDetails: " + ex.getMessage());
        }
    }
    
    /**
     * Send desired data to JMixNet server.
     * 
     * @param data Data to be sent.
     * @throws JMixNetException If data lenght is too long.
     */
    public void sendData(byte[] data, int dataLength) throws JMixNetException {
        //verify data length
        if (dataLength > maxDataSize) {
            throw new DataLengthException("Data length exceeded, maximum allowed is " + maxDataSize + " bytes.");
        }
        //put the address, port, meaningful data size, and meaningful data
        tempBuffer.clear();
        tempBuffer.putShort((short)dataLength);
        tempBuffer.put(data, 0, dataLength);
        
        //verify the need for padding 
        int padding = ((dataDescSize + dataLength) % sessionKeySize > 0) ? 
            sessionKeySize - ((dataDescSize + dataLength) % sessionKeySize) : 0;
        srcBufContentSize = dataDescSize + dataLength + padding;
        
        try {
            int i;
            //take each certificate and build the digital envelopes            
            for (i = 0; i < serverCerts.size(); i++) {
                //define buffers roles
                srcBuf = (i % 2 == 0) ? tempBuffer.array() : dataBuffer.array();
                dstBuf = (i % 2 == 0) ? dataBuffer.array() : tempBuffer.array();
                
                //set the server certificate
                asymCrypto.setCertificate((Certificate)serverCerts.get(i));
                sessionKeyFieldSize = X509CertificateInfo.getKeyBytesSize((X509Certificate)serverCerts.get(i));
    
                //set symmetric key and get encoded parameters            
                symCrypto.generateKey();
                
                //build n-1 envelopes 
                if (i < serverCerts.size() - 1) {
                    //verify if this is the one before the last 
                    if (i < serverCerts.size() - 2) {
                        //encrypt the source
                        symCrypto.encrypt(srcBuf, 0, srcBufContentSize, dstBuf, sessionKeyFieldSize);
                        srcBufContentSize += sessionKeyFieldSize;

                        //encrypt simmetric key and cipher parameters
                        asymCrypto.addDataToEncrypt(symCrypto.getEncodedKey(), 0, symCrypto.getEncodedKey().length, dstBuf, 0);
                        encodedParameters = symCrypto.getParameters().getEncoded();                        
                        asymCrypto.encrypt(encodedParameters, 0, encodedParameters.length, dstBuf);
                    }
                    //the one before the last needs to take into account the hash code to the last envelope
                    else {
                        //encrypt the source
                        symCrypto.encrypt(srcBuf, 0, srcBufContentSize, dstBuf, sessionKeyFieldSize + hashFieldSize);

                        //encrypt simmetric key
                        asymCrypto.addDataToEncrypt(symCrypto.getEncodedKey(), 0, symCrypto.getEncodedKey().length, dstBuf, 0);
                        encodedParameters = symCrypto.getParameters().getEncoded();
                        asymCrypto.encrypt(encodedParameters, 0, encodedParameters.length, dstBuf, hashFieldSize);
                    }
                }
                //build the last envelope
                else {
                    //calculate the hash of the source (with padding) and add it to the begining of the source
                    digester.update(srcBuf, hashFieldSize, messageSize - sessionKeyFieldSize - hashFieldSize);
                    digester.digest(srcBuf, 0, hashFieldSize);
                    
                    //encrypt the source
                    symCrypto.encrypt(srcBuf, 0, messageSize - sessionKeyFieldSize, dstBuf, sessionKeyFieldSize);

                    //encrypt simmetric key
                    asymCrypto.addDataToEncrypt(symCrypto.getEncodedKey(), 0, symCrypto.getEncodedKey().length, dstBuf, 0);
                    encodedParameters = symCrypto.getParameters().getEncoded();
                    asymCrypto.encrypt(encodedParameters, 0, encodedParameters.length, dstBuf);
                }
                if (encodedParameters.length != 18) System.err.println("Parameter Length diff from 18");
            }
            //send the message
            serverOutput.write(dstBuf);
        }
        catch (Exception ex) {
            throw new JMixNetException(ex);
        } 
    }
    
    /**
     * When running the client directly, with no application tayloring, this main method is used.
     * 
     * @param args Config file name, when "jmixnet.cfg" is not supposed to be used. 
     * @throws JMixNetException On unrecoverable error.
     */
    public static void main(String[] args) throws JMixNetException, FileNotFoundException {

        JMixNetClient client;
        ByteBuffer buffer = ByteBuffer.allocate(1024);
        
        //to read data from console
        BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
        String line; 

        //send messages to example server (br.edu.ufsc.labsec.jmixnet.example.ReceiverServer)
        InetAddress receiverAddress = null;
        short receiverPort = 2000;
        try {
            //receiverAddress = InetAddress.getLocalHost();
            receiverAddress = InetAddress.getByName("192.168.128.11");
        }
        catch (UnknownHostException ex) {
            ex.printStackTrace();
        }
        //start client with desired config file        
        if (0 == args.length) 
            client = new JMixNetClient();
        else
            client = new JMixNetClient(new FileInputStream(args[0]));

        System.out.println("Client running.");
        client.connect();
        
        //send lines read from console until "exit" is typed
        /*line = reader.readLine();
        while (!line.equals("exit")) {
            client.sendData((line + "\n").getBytes());
            line = reader.readLine();
        }*/
        
        long lastTime = new Date().getTime();
        long now;
        double elapsedTime;
        int sentMsgs = 0;
        String numero;
        int maxDataSize = client.maxDataSize;
        byte[] buf = new byte[maxDataSize];
        int totalMsgs = 2000;
        int i = 0;
        //for (i = 0; i < totalMsgs; i++) {
        do {
            //verify exit condition
            try {i = System.in.available();} catch (IOException ex) {}
            
            //send message
            numero = (new Integer(i)).toString();
            buffer.clear();
            buffer.put(receiverAddress.getAddress());
            buffer.putShort(receiverPort);
            buffer.put(numero.getBytes());
            client.sendData(buffer.array(), numero.getBytes().length + 6);
            //client.sendData(buf, maxDataSize);
            sentMsgs++;

            //calculate throughput
            now = new Date().getTime();
            elapsedTime = (now - lastTime) / 1000.0;
            if (elapsedTime > 5) { 
                lastTime = now;
                System.out.println("Speed: " + sentMsgs / elapsedTime + " msg/s.");
                sentMsgs = 0;
            }
        } while (i == 0);
        client.disconnect();
    }
}