/******************************************************************************

JMixNet - A Java MixNet Implementation

Fabiano Castro Pereira
LabSEC - Computer Security Laboratory (www.labsec.ufsc.br)
Informatics and Statistics Department (www.inf.ufsc.br)
Federal University of Santa Catarina  (www.ufsc.br)  
Florianopolis, Santa Catarina, Brazil

JMixNet is a registered trademark of the Federal University of Santa Catarina.
No commercial use of this trademark may be made without prior written
permission of the Federal University of Santa Catarina.\

******************************************************************************/
package br.edu.ufsc.labsec.jmixnet;

import br.edu.ufsc.labsec.util.*;
import br.edu.ufsc.labsec.crypto.*;

import java.io.InputStream;
import java.net.*;
import java.util.*;

/**
 * @author castro
 *
 * JMixNet configuration manager.
 */
public class JMixNetConfigManager extends ConfigManager {
    
    protected static int DEFAULT_PORT = 1981;
    protected static int DEFAULT_MSG_SIZE = 8192;
    protected static int DEFAULT_MSG_BATCH_SIZE = 100;
    protected static int DEFAULT_POOL_SIZE = 50;
    protected static int DEFAULT_FLUSH_PERCENT = 70;
    protected static int DEFAULT_FLUSH_PERIOD = 500;
    protected static int DEFAULT_MAX_TRACK_SIZE = 100;
    
    protected int port;
    protected String address;
    protected int messageSize = DEFAULT_MSG_SIZE;
    protected int messageBatchSize = DEFAULT_MSG_BATCH_SIZE;
    protected int poolSize = DEFAULT_POOL_SIZE;
    protected int flushPercent = DEFAULT_FLUSH_PERCENT;
    protected int flushPeriod = DEFAULT_FLUSH_PERIOD;
    protected int maxMsgTrackSize = DEFAULT_MAX_TRACK_SIZE;
    protected String previousServer = null;
    protected String nextServer = null;
    protected int nextServerPort = DEFAULT_PORT;
    protected String msgDelivererClass;
    
    public JMixNetConfigManager() throws ConfigException {
        super("jmixnet.cfg");
        readConfiguration();
    }

    public JMixNetConfigManager(InputStream config) throws ConfigException {
        super(config);
        readConfiguration();
    }

    protected void readConfiguration() throws ConfigException {
        String chain;
        
        //get the port to be used, default is DEFAULT_PORT
        port = getInt("port", DEFAULT_PORT);
        
        //get the address to be used (dot-quadded or fully qualified), default is canonical hostname
        try {
            address = getProperty("address", InetAddress.getLocalHost().getCanonicalHostName());
        }
        catch (UnknownHostException ex) {
            ex.printStackTrace();
        }
        
        //get the message size to be used, default is DEFAUL_MSG_SIZE
        messageSize = getInt("messageSize", DEFAULT_MSG_SIZE);
        //get the message batch size to be used, default is DEFAULT_MSG_BATCH_SIZE
        messageBatchSize = getInt("messageBatchSize", DEFAULT_MSG_BATCH_SIZE);
        //get the message pool size to be used, default is DEFAULT_POOL_SIZE
        poolSize = getInt("poolSize", DEFAULT_POOL_SIZE);
        //get the flush percent to be used, default is DEFAULT_FLUSH_PERCENT
        flushPercent = getInt("flushPercent", DEFAULT_FLUSH_PERCENT);
        //get the flush period to be used, default is DEFAULT_FLUSH_PERIOD
        flushPeriod = getInt("flushPeriod", DEFAULT_FLUSH_PERIOD);
        //get the max message track size 
        maxMsgTrackSize = getInt("maxMsgTrackSize", DEFAULT_MAX_TRACK_SIZE);
        //get the message deliverer class, default is JMixNetDeliverer 
        msgDelivererClass = getString("msgDelivererClass", "br.edu.ufsc.labsec.jmixnet.JMixNetDeliverer");
        
        //get the chain to define the previous and next servers
        chain = getProperty("chain");
        if (chain == null) throw new MissingConfigException("chain");
        
        //list all chain addresses
        String[] addresses = chain.split(",");
        String currentAddress = null;
        String[] addressPort = null;
        
        //seek for local address
        for (int i = 0; i < addresses.length; i++) {
            //split the address part (before the colon) and the port part (after the colon)
            addressPort = addresses[i].split(":");
            
            if (//local address was found?
                (addressPort[0].equals(address)) &&
                //config port is not set or is the same?
                (addressPort.length == 1 || (port == Integer.parseInt(addressPort[1]))) ) 
            {
                //the previous server is the address of the previous iteration
                previousServer = currentAddress;
                
                //verify if it is not the last
                if (i < addresses.length - 1) {
                    //set the next server info
                    addressPort = addresses[i + 1].split(":");
                    nextServer = addressPort[0];
                    if (addressPort.length > 1)
                        nextServerPort = Integer.parseInt(addressPort[1]);
                }
                break;
            }
            currentAddress = addressPort[0];
        }
        //verify that previous or next servers was set
        if ((previousServer == null) && (nextServer == null)) {
            throw new WrongConfigException("chain");
        }
    }
    
    public String getAddress() {
        return address;
    }
    
    public int getPort() {
        return port;
    }

    public String getNextServer() {
        return nextServer;
    }
    
    public int getNextServerPort() {
        return nextServerPort;
    }
    
    public String getPreviousServer() {
        return previousServer;
    }
    
    public int getMessageSize() {
        return messageSize;
    }
    
    public int getMessageBatchSize() {
        return messageBatchSize;
    }
    
    public int getPoolSize() {
        return poolSize;
    }
    
    public int getFlushPercent() {
        return flushPercent;
    }
    
    public int getFlushPeriod() {
        return flushPeriod;
    }
    
    public String getMsgDelivererClass() {
        return msgDelivererClass;
    }
    
    public String getKeyStoreLocation() {
        return getProperty("keyStoreLocation", "JMixNetServer.keystore");
    }
    
    public String getKeyStorePassword() {
        return getProperty("keyStorePassword", "jmixnet");
    }
    
    public Collection getServerCerts() throws ConfigException {
        try {
            return X509CertificateGenerator.genCertificates(
                JMixNetConfigManager.class.getClassLoader().getResourceAsStream("serverCertificates.b64"));
        }
        catch (CryptoException ex) {
            throw new ConfigException("Error generating certificates.", ex);
        }
    }

    public int getMaxMsgTrackSize() {
        return maxMsgTrackSize;
    }

    public HashMap getAllowedClients() {
        HashMap allowedSet = new HashMap();
        String property = getProperty("allowedClients"); 
        if (property != null) {
            String[] allowedClients = property.split(",");
            int numAllowed = allowedClients.length;
            for (int i = 0; i < numAllowed; i++) {
                allowedSet.put(allowedClients[i], null);
            }
        }
        return allowedSet;
    }
}