/******************************************************************************

JMixNet - A Java MixNet Implementation

Fabiano Castro Pereira
LabSEC - Computer Security Laboratory (www.labsec.ufsc.br)
Informatics and Statistics Department (www.inf.ufsc.br)
Federal University of Santa Catarina  (www.ufsc.br)  
Florianopolis, Santa Catarina, Brazil

JMixNet is a registered trademark of the Federal University of Santa Catarina.
No commercial use of this trademark may be made without prior written
permission of the Federal University of Santa Catarina.\

******************************************************************************/
package br.edu.ufsc.labsec.jmixnet;

import br.edu.ufsc.labsec.crypto.*;
import br.edu.ufsc.labsec.util.ConfigException;

import java.net.*;
import javax.net.ssl.*;

import org.bouncycastle.jce.provider.BouncyCastleProvider;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;

import java.security.*;
import java.security.cert.*;

import java.util.*;

/**
 * @author castro
 *
 * Server to accept new connections
 */
public class JMixNetServer extends Thread {

    //configuration-related
    protected JMixNetConfigManager config;

    //message handling
    protected int messageSize;
    protected MessageBatch messageBatch;
    protected int sessionKeyFieldSize;
    protected final int sessionKeySize = 16; //AES
    protected final int parametersSize = 18; //AES parameters 
    protected final int hashFieldSize = 16; //MD5
    protected byte[] receivedHash = new byte[hashFieldSize];
    protected String strReceivedHash;
    protected byte[] plainData;
    protected Iterator iterator;
    protected int flushPeriod;
    protected long lastFlush;
    protected boolean isGoodMessage;
    protected HashMap receivedMsgTrack = new HashMap();
    protected int maxMsgTrackSize;
    
    //accounting
    protected int connectedClients = 0;
    protected int refusedConnections;
    protected int badMsgCount;
    protected int processedMsgCount;
    
    //flush-related
    protected ByteBuffer flushMessage;
    protected LinkedList flushList;
    protected DelivererThread delivererThread;
    
    //reports
    protected PrintStream errorReport = System.err;
    protected PrintStream serverReport = System.out;
    
    //cryptographers and hash
    protected SymCryptographer symCrypto;
    protected AsymCryptographer asymCrypto;
    protected MessageDigest digester;
    
    //to wait for previous server
    protected SSLServerSocketFactory serverSocketFactory = null;
    protected SSLServerSocket serverSocket = null;
    protected SSLSocket previousServerSocket = null;
        
    //to connect to the next server        
    protected SSLSocketFactory socketFactory = null;
    protected SSLSocket nextServerSocket = null;
        
    //to receive and send messages
    protected InputStream previousServerInput = null;
    protected OutputStream nextServerOutput = null;
    protected boolean serviceActive = true;
    protected Selector selector = null;

    //to receive client connections
    protected ServerSocketChannel serverChannel = null;
    protected SocketChannel clientSocket = null;
    protected HashMap clientBuffers = new HashMap();
    protected HashMap allowedClients;

    /**
     * Construct the server with initial setup.
     */
    public JMixNetServer() throws JMixNetException {
        try {
            config = new JMixNetConfigManager();
        }
        catch (ConfigException ex) {
            showError("Config error occurred:\n" + ex.getMessage());
            throw new JMixNetException(ex);
        }
        configureServer();
    }
        
    /**
     * Construct the server with initial setup, based on specified config file.
     * 
     * @param configFile Configuration file other than "jmixnet.cfg"
     */
    public JMixNetServer(InputStream configStream) throws JMixNetException {
        try {
            config = new JMixNetConfigManager(configStream);
        }
        catch (ConfigException ex) {
            showError("Config error occurred:\n" + ex.getMessage());
            throw new JMixNetException(ex);
        }
        configureServer();
    }
        
    /**
     * Configurations based on config file.
     */
    protected void configureServer() throws JMixNetException {
        //ensure Bouncy Castle Provider is available
        if (Security.getProvider("BC") == null) {
            Security.addProvider(new BouncyCastleProvider());
        }
        //SSL-related settings
        System.setProperty("javax.net.ssl.keyStore", config.getKeyStoreLocation());
        System.setProperty("javax.net.ssl.keyStorePassword", config.getKeyStorePassword());
        
        //message batch config
        messageSize = config.getMessageSize();
        messageBatch = new MessageBatch(config.getMessageBatchSize(), 
            config.getPoolSize(), config.getFlushPercent(), messageSize);
        flushPeriod = config.getFlushPeriod();
        plainData = new byte[messageSize];
        maxMsgTrackSize = config.getMaxMsgTrackSize();
        
        //define allowed clients
        allowedClients = config.getAllowedClients();
        
        //cryptography services and hash
        try {
            symCrypto = new SymCryptographer();
            asymCrypto = new AsymCryptographer(config.getKeyStoreLocation(), config.getKeyStorePassword());
            asymCrypto.useAlias("serverKey", null);
            sessionKeyFieldSize = X509CertificateInfo.getKeyBytesSize((X509Certificate)asymCrypto.getCertificate()); 
            digester = MessageDigest.getInstance("MD5");
        }
        catch (CryptoException ex) {
            showError("Unable to use cryptography services:\n" + ex.getMessage());
            throw new JMixNetException(ex);
        }
        catch (NoSuchAlgorithmException ex) {
            showError("Unable to get a message digest instance:\n" + ex.getMessage());
            throw new JMixNetException(ex);
        }
    }
    
    /**
     * Uses error report to display desired error message.
     * 
     * @param message Error message to be shown.
     */
    protected void showError(String message) {
        errorReport.println("ERROR: " + message);
    }
    
    /**
     * Uses server report to display desired log message.
     * 
     * @param message Log message to be shown.
     */
    protected void showLog(String message) {
        serverReport.println("Log: " + message);
    }
    
    /**
     * Set the PrintStream object to be used as the error report.
     * 
     * @param report PrintStream object to be used as the error report.
     */
    public void setErrorReport(PrintStream report) {
        errorReport = report;
    }

    /**
     * Set the PrintStream object to be used as the server report.
     * 
     * @param report PrintStream object to be used as the server report.
     */
    public void setServerReport(PrintStream report) {
        serverReport = report;
    }

    /**
     * Based on the mixnet chain configuration, connects to the next mix.
     */
    protected void connectToNextServer() throws JMixNetException {
        showLog("Connecting to next server on port " + String.valueOf(config.getNextServerPort()) + "...");
        try {
            //stablish connection
            nextServerSocket = (SSLSocket)socketFactory.createSocket(config.getNextServer(), config.getNextServerPort());
            nextServerOutput = nextServerSocket.getOutputStream();
            showLog("Connected to next server.");

            //starts deliverer thread
            delivererThread = new DelivererThread();
            delivererThread.setServerOutput(nextServerOutput);
            delivererThread.start();
        }
        catch (UnknownHostException ex) {
            showError("Unable to connect to the next server. The host \"" + config.getNextServer() + "\" is unknown.");
            throw new JMixNetException(ex);
        }
        catch (ConnectException ex) {
            showError("The next server is not available:\n" + ex.getMessage());
            throw new JMixNetException(ex);
        }
        catch (IOException ex) {
            showError("An IO error occurred when trying to connect to the next server:\n" + ex.getMessage());
            throw new JMixNetException(ex);
        }
    }
    
    /**
     * Based on the mixnet chain configuration, waits for the connection of the previous mix.
     */
    protected void waitPreviousServer() throws JMixNetException {
        try {
            boolean connected = false;
            while (!connected) {
                showLog("Waiting on port " + config.getPort() + " for previous server connection...");
                previousServerSocket = (SSLSocket)serverSocket.accept();
                
                //verify if the connected peer is the desired one
// UNDO                if (previousServerSocket.getSession().getPeerHost().equals(config.getPreviousServer())) {
                    connected = true;
                    showLog("Previous server connected.");
                    previousServerInput = previousServerSocket.getInputStream();
/*                }
                else {
                    showLog("Refused connection from '" + previousServerSocket.getSession().getPeerHost() + "'.");
                    previousServerSocket.close();
                }*/
            }
        }
        catch (IOException ex) {
            showError("Error waiting for previous server connection:\n" + ex.getMessage());
            throw new JMixNetException(ex);
        }
    }
    
    /**
     * Get the messages ready to be delivered and pass them to the delivererThread thread.
     * 
     * @param msgList The messages to be delivered.
     */
    protected void deliverMessages(LinkedList msgList) {
        //set new messages to be delivered and get the handled buffers
        msgList = delivererThread.exchangeBuffers(msgList);
        messageBatch.addEmptyBuffers(msgList);
        
        //free
        msgList.clear();
        msgList = null;
    }

    /**
     * Verifies if a new client message is repeated, discarding it (if true), 
     * or adding this to the message batch (if false).
     * This is used only by the first server.
     * 
     * @param newMsg Message just received.
     */
    protected void handleClientMessage(ByteBuffer newMsg) {
        isGoodMessage = true;
        try {
            //take the session key and cipher parameters (uses plainData as temporary buffer)
            asymCrypto.decrypt(newMsg.array(), 0, sessionKeyFieldSize, plainData);
            symCrypto.setEncodedKey(plainData, 0, sessionKeySize);
            symCrypto.setEncodedParameters(plainData, sessionKeySize, parametersSize);
            
            //decrypt data
            symCrypto.decrypt(newMsg.array(), sessionKeyFieldSize, messageSize - sessionKeyFieldSize, plainData);
            
            //compute and compare hashes
            digester.update(plainData, hashFieldSize, messageSize - sessionKeyFieldSize - hashFieldSize);
            System.arraycopy(plainData, 0, receivedHash, 0, hashFieldSize);
            isGoodMessage = MessageDigest.isEqual(digester.digest(), receivedHash);
            
            //verify message repeat
            strReceivedHash = new String(receivedHash);
            if (receivedMsgTrack.containsKey(strReceivedHash)) {
                showLog("Repeated message received.");
                badMsgCount++;
                isGoodMessage = false;
            }
            else {
                if (receivedMsgTrack.size() > maxMsgTrackSize) {
                    receivedMsgTrack.clear();
                }
                receivedMsgTrack.put(strReceivedHash, null);
            }
        }
        catch (Exception ex) {
            showError("Error handling client message.\n" + ex.getMessage());
            isGoodMessage = false;
        }
        newMsg.clear();
        if (isGoodMessage) {
            //leave the message ready to be sent to the next server
            newMsg.put(plainData, hashFieldSize, plainData.length - hashFieldSize).clear();
            messageBatch.addFullBuffer(newMsg);
            processedMsgCount++;

            //verify flush condition
            if (messageBatch.isReadyToFlush() && ((new Date()).getTime() - lastFlush > flushPeriod)) {
                deliverMessages(messageBatch.getFlushList());
                lastFlush = (new Date()).getTime(); 
            }
        }
        else {
            //treat the bad message as an empty buffer
            messageBatch.addEmptyBuffer(newMsg);
            badMsgCount++;
            showLog("Bad message received.");
        }
    }

    /**
     * Decrypts the message and add it to the message batch.
     * This is used only by middle or last servers.
     * 
     * @param newMsg Message just received.
     */
    protected void handleServerMessage(ByteBuffer newMsg) {
        try {
            //take the session key and cipher parameters (uses plainData as temporary buffer)
            asymCrypto.decrypt(newMsg.array(), 0, sessionKeyFieldSize, plainData);
            symCrypto.setEncodedKey(plainData, 0, sessionKeySize);
            symCrypto.setEncodedParameters(plainData, sessionKeySize, parametersSize);            

            //decrypt data and put the plain data into the plaindata array
            symCrypto.decrypt(newMsg.array(), sessionKeyFieldSize, messageSize - sessionKeyFieldSize, plainData);
        
            //leave the message ready to be sent
            newMsg.clear();
            newMsg.put(plainData).clear();
            messageBatch.addFullBuffer(newMsg);
        
            //verify flush condition
            if (messageBatch.isReadyToFlush() && ((new Date()).getTime() - lastFlush > flushPeriod)) {
                //flush
                deliverMessages(messageBatch.getFlushList());
                lastFlush = (new Date()).getTime(); 
            }
            processedMsgCount++;
        }
        catch (Exception ex) {
            messageBatch.addEmptyBuffer(newMsg);
            badMsgCount++;
            //UNDO showError("Error handling server message.\n" + ex.getMessage());
        }
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run()
     */
    public void run() {
        ByteBuffer msgBuffer;
                
        showLog("Server started.");
        try {
            //connect to next server if this is not the last
            if (config.getNextServer() != null) {
                socketFactory = (SSLSocketFactory)SSLSocketFactory.getDefault();
                connectToNextServer();
            }
            //verify if this is the last, or a middle server
            if (config.getPreviousServer() != null) {
                
                //open port and wait for previous server
                serverSocketFactory = (SSLServerSocketFactory)SSLServerSocketFactory.getDefault();
                try {
                    serverSocket = (SSLServerSocket)serverSocketFactory.createServerSocket(config.getPort());
                }
                catch (IOException ex) {
                    showError("Unable to create server socket:\n" + ex.getMessage());
                    throw new JMixNetException(ex);
                }
                waitPreviousServer();
                
                //verify server position
                if (config.getNextServer() == null) {
                    //create delivererThread thread
                    delivererThread = new DelivererThread(config.getMsgDelivererClass());
                    delivererThread.start();
                    showLog("MixNet service active as the LAST node.");
                }
                else {
                    showLog("MixNet service active as a MIDDLE node.");
                }
                
                boolean messageReceived;
                int amountRead;
                int totalRead;

                //wait for messages from previous server while service is active
                while (serviceActive) {
                    messageReceived = false;
                    totalRead = 0;
                    msgBuffer = messageBatch.getEmptyBuffer();
                    
                    //wait whole message
                    while (!messageReceived) {
                        try {
                            amountRead = previousServerInput.read(msgBuffer.array(), totalRead, messageSize - totalRead);
                        }
                        catch (IOException ex) {
                            //verify if this was a connection reset or an error 
                            if (ex.getMessage().equals("Connection reset")) {
                                showLog("Previous server disconnected.");
                            }
                            else {
                                showError("Error reading from previous server socket:\n" + ex.getMessage());
                            }
                            
                            //clear buffer
                            msgBuffer.clear();
                            messageBatch.addEmptyBuffer(msgBuffer);
                            
                            waitPreviousServer();
                            break;
                        }
                        totalRead += amountRead;
                        messageReceived = totalRead == messageSize;
                    }
                    //handle message
                    if (messageReceived) {
                        handleServerMessage(msgBuffer);
                    }                    
                }
                try {                    
                    serverSocket.close();
                }
                catch (IOException ex) {
                    showError("Unable to close server socket:\n" + ex.getMessage());
                }
            }
            //wait for client connections, as we are the first server
            else {
                SocketChannel channel = null;
                SelectionKey key = null;
                Iterator it = null;
                
                //open a non-blocking service
                try {
                    serverChannel = ServerSocketChannel.open();
                    serverChannel.socket().bind(new InetSocketAddress(config.getPort()));
                    serverChannel.configureBlocking (false);
                    selector = Selector.open();
                }
                catch (IOException ex) {
                    if (null == serverChannel)
                        showError("Unable to open server channel:\n" + ex.getMessage());
                    else if (!serverChannel.socket().isBound())
                        showError("Unable to listen for clients. Por already in use: " + String.valueOf(config.getPort()));
                    throw new JMixNetException(ex);                    
                }
                int amountRead;
                
                //accept new connections thru selector
                serverChannel.register(selector, SelectionKey.OP_ACCEPT);
                
                //keep waiting while the service is active
                showLog("MixNet service active as the FIRST node.");
                while (serviceActive) {
                    //if we have nothing to do, back and test if service is still active
                    try {
                        if (selector.select() == 0) {
                            continue;
                        }
                    }
                    catch (IOException ex) {
                        showError("Unable to select channels:\n" + ex.getMessage());
                    }
                    it = selector.selectedKeys().iterator();
                    
                    //handle each key
                    while (it.hasNext()) {
                        key = (SelectionKey) it.next();
                        
                        //verify if this is a new connection
                        if (key.isAcceptable()) {
                            try {
                                channel = serverChannel.accept();
                            }
                            catch (IOException ex) {
                                showError("Unable to accept client connection:\n" + ex.getMessage());
                            }
                            //verify if connection is still active
                            if (channel != null) {
                                //allow only config file listed IPs
                                if ((allowedClients.size() > 0) && !allowedClients.containsKey(channel.socket().getInetAddress().getHostAddress())) {
                                    try {
                                        showLog("Connection from '" + channel.socket().getInetAddress().getCanonicalHostName() + "' refused.");
                                        refusedConnections++;
                                        channel.close();
                                    }
                                    catch (IOException ex) {
                                        showError("Unable to close client channel:\n" + ex.getMessage());
                                    }
                                }
                                //allow only one active connection per IP address
                                else if (!clientBuffers.containsKey(channel.socket().getInetAddress())) {
                                    try {
                                        //define the buffer this client will write to
                                        clientBuffers.put(channel.socket().getInetAddress(), messageBatch.getEmptyBuffer());
                                        
                                        //setup channel
                                        channel.configureBlocking (false);
                                        channel.register(selector, SelectionKey.OP_READ);
                                        showLog("Client '" + channel.socket().getInetAddress().getCanonicalHostName() + "' connected.");
                                        connectedClients++;
                                    }
                                    catch (IOException ex) {
                                        showError("Unable to set new connection channel to non-blocking:\n" + ex.getMessage());
                                    }
                                }
                                else {
                                    try {
                                        showLog("Extra client connection from '" + channel.socket().getInetAddress().getCanonicalHostName() + "' refused.");
                                        refusedConnections++;
                                        channel.close();
                                    }
                                    catch (IOException ex) {
                                        showError("Unable to close client channel:\n" + ex.getMessage());
                                    }
                                }
                            }
                        }
                        //verify if this is data to be read
                        if (key.isReadable()) {
                            channel = (SocketChannel) key.channel();
                            msgBuffer = (ByteBuffer) clientBuffers.get(channel.socket().getInetAddress());
                            
                            //after a whole message is received, the buffer must be taken again
                            if (msgBuffer == null) {
                                msgBuffer = messageBatch.getEmptyBuffer();
                                clientBuffers.put(channel.socket().getInetAddress(), msgBuffer);
                            }
                            try {
                                if ((amountRead = channel.read(msgBuffer)) > 0) {
                                    //verify if the whole message was read
                                    if (msgBuffer.position() == messageSize) {
                                        //handle message
                                        msgBuffer.flip();
                                        handleClientMessage(msgBuffer);
                                        clientBuffers.remove(channel.socket().getInetAddress());
                                    }
                                }
                                if (amountRead < 0) {
                                    //on EOF, remove client and close channel
                                    showLog("Client '" + channel.socket().getInetAddress().getCanonicalHostName() + "' disconnected.");
                                    connectedClients--;
                                    clientBuffers.remove(channel.socket().getInetAddress());
                                    channel.close();
                                }          
                            }
                            catch (IOException ex) {
                                showError("Unable to communicate with client:\n" + ex.getMessage());
                                showLog("Client '" + channel.socket().getInetAddress().getCanonicalHostName() + "' disconnected.");
                                connectedClients--;
                                clientBuffers.remove(channel.socket().getInetAddress());
                                try {channel.close();} 
                                catch (IOException e) {
                                    showError("Unable to close client channel:\n" + ex.getMessage());
                                }
                            }
                            
                        }
                        it.remove( );
                    }
                }
            }
            showLog("MixNet service finished.");
            
        } 
        catch (ClosedByInterruptException ex) {
            showLog("MixNet service finished through interruption.");
        } 
        catch (ClosedChannelException ex) {
            ex.printStackTrace();
        }
        catch (JMixNetException ex) {
            showError(ex.getMessage());
        }
    }
    
    /**
     * Finishes the service.
     */
    public void finishService() {
        //flush remaining messages
        deliverMessages(messageBatch.getWholeList());
        serviceActive = false;
        
        //verify if we are the first server
        if (config.getPreviousServer() == null)
            selector.wakeup();
            
        this.interrupt();
    }

    public static void main(String[] args) throws JMixNetException, FileNotFoundException {
        JMixNetServer server;
        BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
        
        //start JMixNet server according to command line args
        if (0 == args.length) 
            server = new JMixNetServer();
        else
            server = new JMixNetServer(new FileInputStream(args[0]));
        
        server.start();
        
        try {
            //wait for console command
            String command = reader.readLine();
            while (!command.equals("exit")) {
                //handle command
                System.out.println("Unknown command: [" + command + "]");
                
                //read the next command
                command = reader.readLine();
            }
        }
        catch (IOException ex) {
            ex.printStackTrace();
        }
        server.finishService();

        System.out.println("Main thread finished.");
    }

    /**
     * @return Returns the number of bad messages received.
     */
    public int getBadMsgCount() {
        return badMsgCount;
    }
    
    /**
     * @return Returns the number of connected clientes.
     */
    public int getConnectedClientes() {
        return connectedClients;
    }
    
    /**
     * @return Returns the number of processed messages.
     */
    public int getProcessedMsgCount() {
        return processedMsgCount;
    }
    
    /**
     * @return Returns the number of refused connections.
     */
    public int getRefusedConnections() {
        return refusedConnections;
    }
    
    /**
     * @return Returns the current number of messages stored in the batch.
     */
    public int getStoredMsgCount() {
        return messageBatch.fullBuffers.size();
    }
}