/******************************************************************************

JMixNet - A Java MixNet Implementation

Fabiano Castro Pereira
LabSEC - Computer Security Laboratory (www.labsec.ufsc.br)
Informatics and Statistics Department (www.inf.ufsc.br)
Federal University of Santa Catarina  (www.ufsc.br)  
Florianopolis, Santa Catarina, Brazil

JMixNet is a registered trademark of the Federal University of Santa Catarina.
No commercial use of this trademark may be made without prior written
permission of the Federal University of Santa Catarina.\

******************************************************************************/
package br.edu.ufsc.labsec.jmixnet;

import java.util.*;
import java.nio.*;

/**
 * @author castro
 *
 * Batch to keep messages processed by the server. 
 */
public class MessageBatch {
    
    //buffers-related
    protected LinkedList emptyBuffers;
    protected LinkedList fullBuffers;
    protected LinkedList flushBuffers;
    protected ByteBuffer anEmptyBuffer;
    protected ByteBuffer aFullBuffer;
    protected int allocationCount = 0;
    
    //configuration-related
    protected int msgSize;
    protected int poolSize;
    protected int flushPercent;
    
    //flush-related
    protected int numChosenMsg;
    protected Iterator it;
    protected ByteBuffer currentBuffer;
    protected int i;

    /**
     * Creates the batch.
     * 
     * @param batchSize Size of the batch.
     * @param poolSize Size of the pool of messages to keep on each round.
     * @param flushPercent Percentage of message to be sent on each round.
     * @param msgSize Size of the messages contained by the batch.
     */
    public MessageBatch(int initialBatchSize, int poolSize, int flushPercent, int msgSize) {
        emptyBuffers = new LinkedList();
        fullBuffers = new LinkedList();
        ByteBuffer newMsgBuffer;
        
        this.msgSize = msgSize; 
        this.poolSize = poolSize; 
        this.flushPercent = flushPercent; 

        //create message buffers
        for (int i = 0; i < initialBatchSize; i++) {
            newMsgBuffer = ByteBuffer.allocate(msgSize);
if (++allocationCount % 20 == 0) System.out.println("CONTRUCTOR up to now: " + allocationCount); //UNDO            
            emptyBuffers.add(newMsgBuffer);
        }
    }
    
    /**
     * Returns an empty buffer.
     */
    public ByteBuffer getEmptyBuffer() {
        if (emptyBuffers.size() > 0) {
            //take from emptyBuffers
            it = emptyBuffers.iterator();
            anEmptyBuffer = (ByteBuffer)it.next();
            it.remove();
        }
        else {
            //allocate new
            anEmptyBuffer = ByteBuffer.allocate(msgSize);
if (++allocationCount % 20 == 0) System.out.println("GETEMPTY up to now: " + allocationCount); //UNDO            
        }        
        return anEmptyBuffer;
    }
    
    /**
     * Add the new buffer to the full buffers set.
     * 
     * @param newBuffer The buffer to be added.
     */
    public void addFullBuffer(ByteBuffer newBuffer) {
        fullBuffers.add(newBuffer);
    }
    
    /**
     * Add the new buffer to the empty buffers set.
     * 
     * @param newBuffer The buffer to be added.
     */
    public void addEmptyBuffer(ByteBuffer newBuffer) {
        emptyBuffers.add(newBuffer);
    }
    
    /**
     * Add all buffers passed in to the empty buffers set.
     * 
     * @param buffers The buffers to be added.
     */
    public void addEmptyBuffers(Collection buffers) {
        emptyBuffers.addAll(buffers);
    }
    
    /**
     * Verify if there are enough messages to make a flush.
     */
    public boolean isReadyToFlush() {
        //if there is at least one message over the pool size, it is ready to flush
        return fullBuffers.size() - poolSize >= 1; 
    }
    
    /**
     * Returns the selected messages to be flushed.
     */
    public LinkedList getFlushList() {
        flushBuffers = new LinkedList();
        
        //number of chosen messages to be flushed
        numChosenMsg = (new Double(flushPercent * (fullBuffers.size() - poolSize) / 100.0)).intValue();
        
        //at least one message must be flushed
        if (numChosenMsg == 0) numChosenMsg = 1;  
        
        //mix and leave just the chosen messages 
        Collections.shuffle(fullBuffers);
        Object obj;
        for (i = 0; i < numChosenMsg; i++) {
            obj = fullBuffers.removeFirst();
            flushBuffers.add(obj);
        }
        return flushBuffers;         
    }

    /**
     * Get all full buffers.
     * 
     * @return all full buffers.
     */
    public LinkedList getWholeList() {
        return new LinkedList(fullBuffers);
    }
}